import numpy as np

class Optimizer():
	def __init__(self, net, batch_size, nb_epochs=None, nb_iterations=None):
		self.net = net
		self.batch_size = batch_size
		self.nb_epochs = nb_epochs
		self.nb_iterations = nb_iterations
		self.regularizers = []

	def compute_gradients(self, m_b):
		return self.net.compute_minibatch_grad(m_b)

	def apply_regularizers(self, grads):
		for regularizer in self.regularizers:
			grads = regularizer.regularize(grads)
		return grads

	def compute_update(self, grads):
		return grads

	def apply_update(self, grads):
		self.net.set_gradients(grads)
		self.net.update_model()

	def train_step(self, m_b):
		grads, loss_m_b = self.compute_gradients(m_b)
		grads = self.apply_regularizers(grads)
		grads = self.compute_update(grads)
		self.apply_update(grads)
		return loss_m_b

	def epoch_loop(self, training_data):
		nb_training_examples = training_data[1][0][0].shape[0]
		indexes = np.arange(nb_training_examples)
		for j in range(self.nb_epochs):
			np.random.shuffle(indexes)
			for i in range(0,nb_training_examples, self.batch_size):
				m_b = ([[input_[indexes[i:i+self.batch_size],:] for input_ in time_step] for time_step in training_data[0]], 
					[[output_[indexes[i:i+self.batch_size],:] for output_ in time_step] for time_step in training_data[1]])
				yield m_b, i+self.batch_size >= nb_training_examples

	def iteration_loop(self, training_data):
		nb_training_examples = training_data[1][0][0].shape[0]
		for i in range(self.nb_iterations):
			indexes = np.random.randint(nb_training_examples, size=self.batch_size)
			m_b = ([[input_[indexes,:] for input_ in time_step] for time_step in training_data[0]], 
					[[output_[indexes,:] for output_ in time_step] for time_step in training_data[1]])
			yield m_b, False

	def iteration_stop_criteria(self, training_data):
		nb_training_examples = training_data[1][0][0].shape[0]
		while self.stop_criteria(self):
			indexes = np.random.randint(nb_training_examples, size=self.batch_size)
			m_b = ([[input_[indexes,:] for input_ in time_step] for time_step in training_data[0]], 
					[[output_[indexes,:] for output_ in time_step] for time_step in training_data[1]])
			yield m_b, False

	def fit(self, training_data,
				func_ep = lambda *x: print(x[0],"training loss:", x[2]),
				func_mb = None):
		if self.nb_iterations and func_mb == None:
			func_mb=func_ep
		self.i_it = self.i_ep = self.loss_ac = 0
		nb_training_examples = training_data[1][0][0].shape[0]
		for m_b, do_func_ep in (self.iteration_loop if self.nb_iterations != None else self.epoch_loop)(training_data):
			loss = self.train_step(m_b)
			self.loss_ac += loss
			if func_mb is not None:
				func_mb(self.i_it, self, loss/m_b[1][0][0].shape[0])
				self.i_it += 1
			if do_func_ep and func_ep is not None:
				func_ep(self.i_ep, self, self.loss_ac/nb_training_examples)
				self.i_ep += 1
				self.loss_ac = 0

	def add_regularizer(self, regularizer):
		assert isinstance(regularizer, Regularizer)
		self.regularizers.append(regularizer)


class SGD(Optimizer):
	def __init__(self, lr_start, lr_end, clipping = None, **kwargs):
		super(SGD, self).__init__(**kwargs)
		self.lr_start = lr_start
		self.lr_end = lr_end
		self.clipping = clipping
		self.lr = self.lr_start
		self.dec_rate = 1

	def compute_update(self, grads):
		grads = [grad*self.lr for grad in grads]
		self.lr *= self.dec_rate
		return grads

	def fit(self, training_data, *args, **kwargs):
		nb_training_examples = training_data[1][0][0].shape[0]
		total_batches = self.nb_iterations if self.nb_iterations != None else self.nb_epochs * np.ceil(nb_training_examples/self.batch_size)
		self.dec_rate = 1
		if total_batches != 1:
			self.dec_rate = (self.lr_end/self.lr_start)**(1/(total_batches-1))
		super(SGD, self).fit(training_data, *args, **kwargs)


class Adam(Optimizer):
	def __init__(self, alpha=0.001, beta1=0.9, beta2=0.999, epsilon=1e-8, **kwargs):
		super(Adam, self).__init__(**kwargs)
		self.alpha = alpha
		self.beta1 = beta1
		self.beta2 = beta2
		self.epsilon = epsilon
		self.net.reset_gradients()
		number_of_gradients = len(self.net.get_gradients())
		self.m = [0]*number_of_gradients 
		self.v = [0]*number_of_gradients 
		self.t = 0
	
	def compute_update(self, grads):
		self.t+=1

		self.m = [self.beta1 * self.m[i] + (1-self.beta1) * grads[i] for i in range(len(grads))]
		self.v = [self.beta2 * self.v[i] + (1-self.beta2) * (grads[i]**2) for i in range(len(grads))]

		m_hat = [m / (1-self.beta1**self.t) for m in self.m]
		v_hat = [v / (1-self.beta2**self.t) for v in self.v]

		grads = [self.alpha * m_hat[i] / (np.sqrt(v_hat[i]) + self.epsilon) for i in range(len(grads))]
		return grads


class Regularizer():
	def __init__(self, net):
		self.net = net

	def regularize(self):
		pass

class Clipping(Regularizer):
	def __init__(self, net, min_, max_):
		super(Clipping, self).__init__(net)
		self.min_ = min_
		self.max_ = max_

	def regularize(self, grads):
		return [np.clip(grad, self.min_, self.max_) for grad in grads]

class L2(Regularizer):
	def __init__(self, net, alpha = 0.0001):
		super(L2, self).__init__(net)
		self.alpha = alpha

	def regularize(self, grads):
		params = self.net.get_params()
		grads = [grads[i] + self.alpha * params[i] for i in range(len(params))]
		return grads

class L1(Regularizer):
	def __init__(self, net, alpha = 0.00001):
		super(L1, self).__init__(net)
		self.alpha = alpha

	def regularize(self, grads):
		params = self.net.get_params()
		grads = [grads[i] + self.alpha * np.sign(params[i]) for i in range(len(params))]
		return grads


































from tkinter import *
from tkinter import ttk, filedialog

from os import listdir
from os.path import isfile, join

import sys
from . import plot


class WBNN_Plots(Tk):
	def __init__(self):
		super(WBNN_Plots, self).__init__()
		self.grid_columnconfigure(0, weight=1)
		self.grid_rowconfigure(0,weight=1)

		self.title("WBNN Plots")

		self.minsize(700, 300)

		self.menu()

		self.file_folders()
		self.plots_side()

	def menu(self):
		menubar = Menu(self)

		# create a pulldown menu, and add it to the menu bar
		filemenu = Menu(menubar, tearoff=0)
		filemenu.add_command(label="Open", command=lambda: self.new_file_folder(filedialog.askdirectory()))
		filemenu.add_command(label="New Plot Area", command=lambda: self.new_plot_area())
		filemenu.add_separator()
		filemenu.add_command(label="Exit", command=self.quit)
		menubar.add_cascade(label="File", menu=filemenu)

		# display the menu
		self.config(menu=menubar)

	def file_folders(self):
		self.file_manager = ttk.Notebook(self, padding=(5, 5, 12, 0))
		self.file_manager.grid_columnconfigure(0, weight=1)
		self.file_manager.grid_rowconfigure(5, weight=1)
		self.file_manager.grid(column=0, row=0, sticky=(N,W,E,S))

		self.folders = []

	def new_file_folder(self, site="."):

		onlyfiles = [f for f in listdir(site) if isfile(join(site, f)) and f.split('_')[0] in ['hist', 'scalar', 'image']]

		onlyfiles = sorted(onlyfiles)

		frame = ttk.Notebook(self, padding=(5, 5, 12, 0))
		frame.grid(column=0, row=0, sticky=(N,W,E,S))
		frame.grid_columnconfigure(0, weight=1)
		frame.grid_rowconfigure(5, weight=1)

		lbl_name = ttk.Label(frame, text="folder:\n"+(site[-49:] if len(site)<=49 else "..."+site[-45:] ))
		lbl_name.grid(column=0, row=0, padx=0, pady=5, sticky=(N,S,E,W))

		c_files = StringVar(value=onlyfiles)

		lbox = Listbox(frame, listvariable=c_files, height=30, width=40)
		lbox.grid(column=0, row=1, rowspan=6, sticky=(N,S,E,W))

		lbox.bind('<<ListboxSelect>>', lambda x: self.append_file_to_plot_area(lbox.curselection()))

		for i in range(1,len(onlyfiles),2):
			lbox.itemconfigure(i, background='#f0f0ff')

		exit_folder = ttk.Button(frame, text='close', command=lambda : self.delete_file_folder(site), default='active')
		exit_folder.grid(column=0, row=20, sticky=S)

		self.file_manager.add(frame, text=str(len(self.folders)+1))

		self.file_manager.select(len(self.folders))

		self.folders.append({
			'folder': site,
			'frame': frame,
			'files': onlyfiles
		})

	def delete_file_folder(self, folder_dir):
		if len(self.folders) == 1:
			return
		idx = -1
		for i, folder in enumerate(self.folders):
			if folder['folder'] == folder_dir:
				idx = i
		self.file_manager.forget(idx)
		del self.folders[idx]
		self.file_manager.select(len(self.folders)-1)

		for i in range(len(self.folders)):
			self.file_manager.tab(self.folders[i]['frame'], text = str(i+1))

	def append_file_to_plot_area(self, file_idx):
		if len(file_idx)==0:
			return
		file_idx = file_idx[0]
		idx_of_folders = self.file_manager.index(self.file_manager.select())
		folder_ = self.folders[idx_of_folders]['folder']
		file_ = self.folders[idx_of_folders]['files'][file_idx]

		idx_of_plots = self.plots_manager.index(self.plots_manager.select())

		if (folder_, file_) in self.plots_confs[idx_of_plots]['files']:
			return

		self.plots_confs[idx_of_plots]['files'].append((folder_, file_))
		self.plots_confs[idx_of_plots]['lbox'].insert('end', file_+" ("+folder_+")" )

		for i in range(1,len(self.plots_confs[idx_of_plots]['files']),2):
			self.plots_confs[idx_of_plots]['lbox'].itemconfigure(i, background='#f0f0ff')


	def plots_side(self):
		self.plots_manager = ttk.Notebook(self, padding=(5, 5, 12, 0))
		self.plots_manager.grid(column=1, row=0, sticky=(N,W,E,S))
		self.plots_manager.grid_columnconfigure(0, weight=1)
		self.plots_manager.grid_rowconfigure(5, weight=1)

		self.plots_confs = []

		self.new_plot_area()

	def new_plot_area(self):

		frame = ttk.Notebook(self, padding=(5, 5, 12, 0))
		frame.grid(column=0, row=0, sticky=(N,W,E,S))
		frame.grid_columnconfigure(0, weight=1)
		frame.grid_rowconfigure(5, weight=1)

		lbox = Listbox(frame, listvariable=None, height=30, width=40)
		lbox.grid(column=0, row=1, rowspan=6, sticky=(N,S,E,W))

		lbox.bind('<<ListboxSelect>>', lambda x: self.remove_file_from_plot_area(lbox.curselection()))

		self.plots_manager.add(frame, text=str(len(self.plots_confs)+1))

		scalar_b = ttk.Button(frame, text='Scalar', command=lambda : self.plot(plot.Scalar), default='active')
		hist_b = ttk.Button(frame, text='Histogram', command=lambda : self.plot(plot.Histogram), default='active')
		image_b = ttk.Button(frame, text='Image', command=lambda : self.plot(plot.Image), default='active')
		scalar_b.grid(column=0, row=20, sticky=(S,W))
		hist_b.grid(column=0, row=20, sticky=S)
		image_b.grid(column=0, row=20, sticky=(S, E))

		exit_plot = ttk.Button(frame, text='close', command=lambda : self.delete_plot_area(frame), default='active')
		exit_plot.grid(column=0, row=22, sticky=S)

		self.plots_confs.append({
			'frame': frame,
			'lbox': lbox,
			'files': []
		})

	def plot(self, plot_class):
		idx_of_plots = self.plots_manager.index(self.plots_manager.select())
		files = [join(folder_, file_) for folder_, file_ in self.plots_confs[idx_of_plots]['files']]
		plot_class(files)

	def delete_plot_area(self, plotter_to_delete):
		if len(self.plots_confs) == 1:
			return
		idx = 0
		for plotter in self.plots_confs:
			if id(plotter['frame']) == id(plotter_to_delete):
				break
			idx += 1
		self.plots_manager.forget(idx)
		del self.plots_confs[idx]
		self.plots_manager.select(len(self.plots_confs)-1)

		for i in range(len(self.plots_confs)):
			self.plots_manager.tab(self.plots_confs[i]['frame'], text = str(i+1))

	def remove_file_from_plot_area(self, remove_index):
		if len(remove_index)==0:
			return
		remove_index = remove_index[0]
		idx_of_plots = self.plots_manager.index(self.plots_manager.select())

		del self.plots_confs[idx_of_plots]['files'][remove_index]

		self.plots_confs[idx_of_plots]['lbox'].delete(remove_index)

		for i in range(1,len(self.plots_confs[idx_of_plots]['files']),2):
			self.plots_confs[idx_of_plots]['lbox'].itemconfigure(i, background='#f0f0ff')
		for i in range(0,len(self.plots_confs[idx_of_plots]['files']),2):
			self.plots_confs[idx_of_plots]['lbox'].itemconfigure(i, background='#ffffff')


def main(args = None):
	if args is None:
		args = sys.argv[1:]

	args = [x for x in args if not isfile(x)]
	if len(args) == 0:
		args.append('.')

	root = WBNN_Plots()
	for folder in args:
		root.new_file_folder(folder)
	root.mainloop()

if __name__ == '__main__':
	main()


import wbnn
import os, time
import numpy as np

from functools import reduce

class Logger():

	def __init__(self, folder):
		self.save_folder = folder
		if not os.path.exists(self.save_folder):
			os.makedirs(self.save_folder)

	def write(self, var_name, val_str):
		with open(os.path.join(self.save_folder, var_name+'.dat'), 'a') as f:
			f.write("%f\t%s\n"%(time.time(), val_str))

	def scalar(self, var_name, value):
		self.write('scalar_'+var_name, str(value))

	def histogram(self, var_name, values, intervals=25):
		min_ = np.min(values)
		max_ = np.max(values)
		pad = (max_-min_)/100
		if pad == 0:
			pad = 0.000001
		min_ -= pad
		max_ += pad
		ys = [0]*intervals
		step = (max_-min_)/(intervals)
		xs = np.arange(min_, max_+step, step).tolist()[:intervals+1]
		for i in range(len(xs)-1):
			ys[i] += np.sum( (values>xs[i]) * (values<=xs[i+1]) )
		self.write('hist_'+var_name, str(step)+'\t'+str(list(zip(xs[:-1],ys))))

	def image(self, var_name, values):
		name = 'image_'+var_name
		times = np.asarray([[time.time()]])
		values = values[np.newaxis,...]
		try:
			data = np.load(os.path.join(self.save_folder, name+".npz"))
			times = np.concatenate([data['times'], times], axis = 0)
			values = np.concatenate([data['values'], values], axis = 0)
		except Exception as e:
			pass
		finally:
			np.savez(os.path.join(self.save_folder, name+".npz"), times=times, values=values)
			# np.save(os.path.join(self.save_folder, name+".npy"), values)



import numpy as np
from . import initializers
import os
import json

class Layer:
	def __init__(self, name):
		self.next = []
		self.prev = []
		self.next_recurrent = []
		self.prev_recurrent = []
		self.name = name
		self.a = None
		self.error = None
		self.b_grad = None
		self.W_grad = None
		self.prop_idx = 0
		self.initializer = initializers.RandomNormal() # Xavier initialization

	def setInitializer(self, initializer):
		assert isinstance(initializer, initializers.Initializer)
		self.initializer = initializer

	def addNext(self, layer):
		self.next.append(layer)
		layer.__addPrev__(self)

	def addNextRecurrent(self, layer):
		self.next_recurrent.append(layer)
		layer.__addPrevRecurrent__(self)

	def __addPrev__(self, layer):
		self.prev.append(layer)

	def __addPrevRecurrent__(self, layer):
		self.prev_recurrent.append(layer)

	def get_Output(self, t=0):
		return self.a[t-1]

	def get_error(self, t=0):
		return self.error[t-1]

	def get_Output_dim(self):
		return self.output_dim

	def get_error_contribution(self, layer, t=0):
		pass

	def backprop_error(self,t=0, external_error=0):
		pass

	def reset_grads(self):
		pass

	def compute_gradients(self):
		pass

	def apply_to_gradients(self, func):
		pass

	def get_params(self):
		pass

	def update(self):
		pass

	def initialize(self):
		pass

	def prop(self, x_labels, proped = []):
		pass

	def reset(self, minibatch_size):
		if len(self.next_recurrent)>0:
			self.a = [np.zeros((minibatch_size, self.get_Output_dim()))]
		else:
			self.a = []
		self.error = []

	def __save__dict__(self):
		pass

	def __load__dict__(self, d):
		pass

	def __dict_maker__(self, d, dir_):
		em = [] if isinstance(d, list) else {}
		if isinstance(d, list):
			for i in range(len(d)):
				if isinstance(d[i], np.ndarray):
					em.append("@"+str(i))
					np.save(dir_+"/@"+str(i), d[i])
				else:
					os.makedirs(dir_+'/'+str(i))
					em.append(self.__dict_maker__(d[i], dir_+'/'+str(i)))
		elif isinstance(d, dict):
			for el in d:
				if isinstance(d[el], np.ndarray):
					em[el] = "@"+str(el)
					np.save(dir_+"/@"+str(el), d[el])
				else:
					os.makedirs(dir_+'/'+str(el))
					em[el] = self.__dict_maker__(d[el], dir_+'/'+str(el))
		else:
			em = d
		return em

	def __dict_load__(d, dir_):
		if isinstance(d, list):
			for i in range(len(d)):
				if isinstance(d[i], str) and d[i][0]=='@':
					d[i] = np.load(dir_+'/'+d[i]+'.npy')
				else:
					d[i] = Layer.__dict_load__(d[i], dir_+'/'+i)
		if isinstance(d, dict):
			for el in d:
				if isinstance(d[el], str) and d[el][0]=='@':
					d[el] = np.load(dir_+'/'+d[el]+'.npy')
				else:
					d[el] = Layer.__dict_load__(d[el], dir_+'/'+el)
		return d

	def save(self, dir_):
		my_id = id(self)
		my_dir = dir_+'/'+str(my_id)
		class_name = self.__class__.__name__
		os.makedirs(my_dir)
		d, ini = self.__save__dict__()
		me = {
			'id': my_id,
			'class': class_name,
			'name': self.name,
			'ini': ini,
			'info': self.__dict_maker__(d, my_dir)
		}
		file_ = json.dumps(me)
		with open(my_dir+"/setup.json", "w") as f:
			f.write(file_)

	def load(dir_):
		with open(dir_+"/setup.json", "r") as f:
			me = json.loads(f.read())
			d = me['info']
			constructor = globals()[me['class']]
			instance = constructor(*me['ini'])
			instance.__load__dict__(Layer.__dict_load__(d, dir_))
			return instance

class Input(Layer):
	def __init__(self, input_dim, name):
		super(Input, self).__init__(name)
		self.input_dim = input_dim

	def get_Output_dim(self):
		return self.input_dim

	def get_params(self):
		return []

	def prop(self, x_labels):
		self.a = self.a + [x_labels]
		return x_labels

	def __save__dict__(self):
		return {}, [self.input_dim, self.name]

class Fully_Connected(Layer):
	def __init__(self, output_dim, name):
		super(Fully_Connected, self).__init__(name)
		self.output_dim = output_dim

	def initialize(self):
		self.W = [self.initializer.get((l.get_Output_dim(), self.get_Output_dim())) for l in self.prev_recurrent + self.prev]
		self.b = self.initializer.get((1, self.output_dim))

	def reset_grads(self):
		self.b_grad = 0
		self.W_grad = [0]*(len(self.prev)+len(self.prev_recurrent))

	def __W_of_layer__(self, layer):
		return self.W[(self.prev_recurrent + self.prev).index(layer)]

	def get_error_contribution(self, layer, t=0):
		return np.dot(
			self.get_error(t=t),
			np.transpose(self.__W_of_layer__(layer))
		)

	def backprop_error(self, t=0):
		#  this function will be used during the BP. to calculate the error
		aux = 0
		if t != 0:
			for layer in self.next_recurrent:
				aux += layer.get_error_contribution(self, t=t+1)
		for layer in self.next:
			aux += layer.get_error_contribution(self, t=t)
		self.error = [aux]+self.error

	def compute_gradients(self):
		num_t = len(self.error)
		b_grad_aux = 0
		W_grad_aux = [0]*len(self.W_grad)
		for t in range(0,-num_t, -1):
			aux = self.get_error(t=t)
			if aux is not 0:
				b_grad_aux += np.sum(aux, axis=0)
				W_grad_aux = list(map(lambda x,y: x+y,
					W_grad_aux,
					[np.dot(np.transpose(l.get_Output(t=t-1)), aux) for l in self.prev_recurrent] +
					[np.dot(np.transpose(l.get_Output(t=t)), aux) for l in self.prev]
				))
		self.b_grad += b_grad_aux
		self.W_grad = list(map(lambda x,y: x+y,
				self.W_grad, W_grad_aux
		))

	def apply_to_gradients(self, func):
		self.b_grad = func(self.b_grad)
		self.W_grad = list(map(func, self.W_grad))

	def get_params(self):
		return [self.b] + self.W

	def update(self):
		self.__update_b__()
		self.__update_W__()

	def __update_W__(self):
		for i in range(len(self.W_grad)):
			self.W[i] -= self.W_grad[i]

	def __update_b__(self):
		self.b -= self.b_grad

	def prop(self):
		inp = 0
		i = 0
		for l in self.prev_recurrent:
			inp += np.dot(l.get_Output(t= 0 if l.prop_idx >= self.prop_idx else -1), self.W[i])
			i+=1
		for l in self.prev:
			inp += np.dot(l.get_Output(t=0), self.W[i])
			i+=1
		out = inp + self.b
		self.a = self.a + [out]
		return out

	def __save__dict__(self):
		return {
			'W': self.W,
			'b': self.b
		}, [self.output_dim, self.name]

	def __load__dict__(self, d):
		self.W = d['W']
		self.b = d['b']

class LSTM(Layer):
	def __init__(self, num_cel, name):
		super(LSTM, self).__init__(name)
		self.name = name
		self.num_cel = num_cel

	def __sigm__(self, z):
		return (np.tanh(z/2)+1)/2

	def __dev_sigm__(self, a):
		return a*(1-a)

	def __tanh__(self, z):
		return np.tanh(z)

	def __dev_tanh__(self, a):
		return 1 - a**2

	def __dev_tanh_z__(self, z):
		return 1 - np.tanh(z)**2

	def get_i(self, t=0):
		return self.i[t-1]

	def get_f(self, t=0):
		return self.f[t-1]

	def get_c(self, t=0):
		return self.c[t-1]

	def get_o(self, t=0):
		return self.o[t-1]

	def get_a_error(self, t=0):
		return self.error_a[t-1]

	def get_o_error(self, t=0):
		return self.error_o[t-1]

	def get_tanh_zc(self, t=0):
		return self.tanh_zc[t-1]

	def get_c_error(self, t=0):
		return self.error_c[t-1]

	def get_f_error(self, t=0):
		return self.error_f[t-1]

	def get_i_error(self, t=0):
		return self.error_i[t-1]
		
	def get_error_contribution(self, layer, t=0):
		idx = self.layers.index(layer)
		return np.dot(self.get_o_error(t=t) * self.__dev_sigm__(self.get_o(t=t)), np.transpose(self.W_o_alprevs[idx]))
		+ np.dot(self.get_c_error(t=t) * self.get_i(t=t) * self.__dev_tanh__(self.get_c(t=t)), np.transpose(self.W_c_alprevs[idx]))
		+ np.dot(self.get_i_error(t=t) * self.__dev_sigm__(self.get_i(t=t)), np.transpose(self.W_i_alprevs[idx]))
		+ np.dot(self.get_f_error(t=t) * self.__dev_sigm__(self.get_f(t=t)), np.transpose(self.W_f_alprevs[idx]))

	def backprop_error(self,t=0, external_error=0):
		"""this function will be used during the BP. to calculate the error"""
		aux = external_error
		# Other layers' contributions
		if t != 0:
			for layer in self.next_recurrent:
				aux += layer.get_error_contribution(self, t=t+1)
		for layer in self.next:
			aux += layer.get_error_contribution(self, t=t)

		if t != 0:
			aux += np.dot(
				self.get_o_error(t=t+1) * self.__dev_sigm__(self.get_o(t=t+1)),
				np.transpose(self.W_o_atprev)
			) + \
			np.dot(
				self.get_c_error(t=t+1) * self.get_i(t=t+1) * self.__dev_tanh__(self.get_c(t=t+1)),
				np.transpose(self.W_c_atprev)
			) + \
			np.dot(
				self.get_i_error(t=t+1) * self.__dev_sigm__(self.get_i(t=t+1)),
				np.transpose(self.W_i_atprev)
			) + \
			np.dot(
				self.get_f_error(t=t+1) * self.__dev_sigm__(self.get_f(t=t+1)),
				np.transpose(self.W_f_atprev)
			)
		
		if isinstance(aux, np.ndarray):
			self.error_a = [aux]+self.error_a

		# error in o
		self.error_o = [self.get_a_error(t=t) * self.__tanh__(self.get_c(t=t))] + self.error_o

		# error in c
		aux = self.w_o_c * self.get_o_error(t=t) * self.__dev_sigm__(self.get_o(t=t)) \
		+ self.get_a_error(t=t) * self.get_o(t=t) * self.__dev_tanh_z__(self.get_c(t=t))
		if t!=0:
			aux += self.get_c_error(t=t+1) * self.get_f(t=t+1) \
			+ self.w_i_ctprev * self.get_i_error(t=t+1) * self.__dev_sigm__(self.get_i(t=t+1)) \
			+ self.w_f_ctprev * self.get_f_error(t=t+1) * self.__dev_sigm__(self.get_f(t=t+1))
		self.error_c = [aux] + self.error_c

		# error in f
		self.error_f = [self.get_c(t=t-1)*self.get_c_error(t=t)] + self.error_f

		# error in i
		self.error_i = [self.get_c_error(t=t)*self.get_tanh_zc(t=t)] + self.error_i

	def reset_grads(self):
		self.W_i_alprevs_grads = [0]*len(self.layers)
		self.W_f_alprevs_grads = [0]*len(self.layers)
		self.W_c_alprevs_grads = [0]*len(self.layers)
		self.W_o_alprevs_grads = [0]*len(self.layers)

		self.W_i_atprev_grads = 0
		self.W_f_atprev_grads = 0
		self.W_c_atprev_grads = 0
		self.W_o_atprev_grads = 0

		self.w_i_ctprev_grads = 0
		self.w_f_ctprev_grads = 0
		self.w_o_c_grads = 0

		self.b_i_grads = 0
		self.b_f_grads = 0
		self.b_c_grads = 0
		self.b_o_grads = 0

	def compute_gradients(self):
		num_t = len(self.error_a)

		W_i_alprevs_grads_aux = [0]*len(self.layers)
		W_f_alprevs_grads_aux = [0]*len(self.layers)
		W_c_alprevs_grads_aux = [0]*len(self.layers)
		W_o_alprevs_grads_aux = [0]*len(self.layers)

		W_i_atprev_grads_aux = 0
		W_f_atprev_grads_aux = 0
		W_c_atprev_grads_aux = 0
		W_o_atprev_grads_aux = 0

		w_i_ctprev_grads_aux = 0
		w_f_ctprev_grads_aux = 0
		w_o_c_grads_aux = 0

		b_i_grads_aux = 0
		b_f_grads_aux = 0
		b_c_grads_aux = 0
		b_o_grads_aux = 0
		for t in range(0,-num_t, -1):
			aux_o = self.get_o_error(t=t) * self.__dev_sigm__(self.get_o(t=t))
			aux_c = self.get_c_error(t=t) * self.get_i(t=t) * self.__dev_tanh__(self.get_c(t=t))
			aux_f = self.get_f_error(t=t) * self.__dev_sigm__(self.get_f(t=t))
			aux_i = self.get_i_error(t=t) * self.__dev_sigm__(self.get_i(t=t))

			for layer in self.layers:
				W_o_alprevs_grads_aux = list(map(lambda x,y: x+y,
					W_o_alprevs_grads_aux,
					[np.dot(np.transpose(l.get_Output(t=t-1)), aux_o) for l in self.prev_recurrent]+\
					[np.dot(np.transpose(l.get_Output(t=t)), aux_o) for l in self.prev]
				))

				W_c_alprevs_grads_aux = list(map(lambda x,y: x+y,
					W_c_alprevs_grads_aux,
					[np.dot(np.transpose(l.get_Output(t=t-1)), aux_c) for l in self.prev_recurrent]+\
					[np.dot(np.transpose(l.get_Output(t=t)), aux_c) for l in self.prev]
				))

				W_f_alprevs_grads_aux = list(map(lambda x,y: x+y,
					W_f_alprevs_grads_aux,
					[np.dot(np.transpose(l.get_Output(t=t-1)), aux_f) for l in self.prev_recurrent]+\
					[np.dot(np.transpose(l.get_Output(t=t)), aux_f) for l in self.prev]
				))

				W_i_alprevs_grads_aux = list(map(lambda x,y: x+y,
					W_i_alprevs_grads_aux,
					[np.dot(np.transpose(l.get_Output(t=t-1)), aux_i) for l in self.prev_recurrent]+\
					[np.dot(np.transpose(l.get_Output(t=t)), aux_i) for l in self.prev]
				))

			W_i_atprev_grads_aux += np.dot(np.transpose(self.get_Output(t=t-1)), aux_i)
			W_f_atprev_grads_aux += np.dot(np.transpose(self.get_Output(t=t-1)), aux_f)
			W_c_atprev_grads_aux += np.dot(np.transpose(self.get_Output(t=t-1)), aux_c)
			W_o_atprev_grads_aux += np.dot(np.transpose(self.get_Output(t=t-1)), aux_o)

			w_i_ctprev_grads_aux += aux_i * self.get_c(t=t-1)
			w_f_ctprev_grads_aux += aux_f * self.get_c(t=t-1)
			w_o_c_grads_aux += aux_o * self.get_c(t=t)

			b_i_grads_aux += aux_i
			b_f_grads_aux += aux_f
			b_c_grads_aux += aux_c
			b_o_grads_aux += aux_o

		self.W_i_alprevs_grads = [self.W_i_alprevs_grads[i] + W_i_alprevs_grads_aux[i] for i in range(len(W_i_alprevs_grads_aux))]
		self.W_f_alprevs_grads = [self.W_f_alprevs_grads[i] + W_f_alprevs_grads_aux[i] for i in range(len(W_f_alprevs_grads_aux))]
		self.W_c_alprevs_grads = [self.W_c_alprevs_grads[i] + W_c_alprevs_grads_aux[i] for i in range(len(W_c_alprevs_grads_aux))]
		self.W_o_alprevs_grads = [self.W_o_alprevs_grads[i] + W_o_alprevs_grads_aux[i] for i in range(len(W_o_alprevs_grads_aux))]

		self.W_i_atprev_grads += W_i_atprev_grads_aux 
		self.W_f_atprev_grads += W_f_atprev_grads_aux 
		self.W_c_atprev_grads += W_c_atprev_grads_aux 
		self.W_o_atprev_grads += W_o_atprev_grads_aux 

		self.w_i_ctprev_grads += np.sum(w_i_ctprev_grads_aux, axis=0)[np.newaxis] 
		self.w_f_ctprev_grads += np.sum(w_f_ctprev_grads_aux, axis=0)[np.newaxis]
		self.w_o_c_grads += np.sum(w_o_c_grads_aux, axis=0)[np.newaxis]

		self.b_i_grads += np.sum(b_i_grads_aux, axis=0)[np.newaxis]
		self.b_f_grads += np.sum(b_f_grads_aux, axis=0)[np.newaxis]
		self.b_c_grads += np.sum(b_c_grads_aux, axis=0)[np.newaxis]
		self.b_o_grads += np.sum(b_o_grads_aux, axis=0)[np.newaxis]

	def apply_to_gradients(self, func):
		self.W_i_alprevs_grads = list(map(func, self.W_i_alprevs_grads))
		self.W_f_alprevs_grads = list(map(func, self.W_f_alprevs_grads))
		self.W_c_alprevs_grads = list(map(func, self.W_c_alprevs_grads))
		self.W_o_alprevs_grads = list(map(func, self.W_o_alprevs_grads))

		self.W_i_atprev_grads = func(self.W_i_atprev_grads)
		self.W_f_atprev_grads = func(self.W_f_atprev_grads)
		self.W_c_atprev_grads = func(self.W_c_atprev_grads)
		self.W_o_atprev_grads = func(self.W_o_atprev_grads)

		self.w_i_ctprev_grads = func(self.w_i_ctprev_grads)
		self.w_f_ctprev_grads = func(self.w_f_ctprev_grads)
		self.w_o_c_grads = func(self.w_o_c_grads)

		self.b_i_grads = func(self.b_i_grads)
		self.b_f_grads = func(self.b_f_grads)
		self.b_c_grads = func(self.b_c_grads)
		self.b_o_grads = func(self.b_o_grads)

	def get_params(self):
		return self.W_i_alprevs + self.W_f_alprevs + self.W_c_alprevs + self.W_o_alprevs +\
			[self.W_i_atprev, self.W_f_atprev, self.W_c_atprev, self.W_o_atprev,
			self.w_i_ctprev, self.w_f_ctprev, self.w_o_c,
			self.b_i, self.b_f, self.b_c, self.b_o]

	def update(self):
		for i in range(len(self.W_i_alprevs)):
			self.W_i_alprevs[i] -= self.W_i_alprevs_grads[i]
			self.W_f_alprevs[i] -= self.W_f_alprevs_grads[i]
			self.W_c_alprevs[i] -= self.W_c_alprevs_grads[i]
			self.W_o_alprevs[i] -= self.W_o_alprevs_grads[i]

		self.W_i_atprev -= self.W_i_atprev_grads
		self.W_f_atprev -= self.W_f_atprev_grads
		self.W_c_atprev -= self.W_c_atprev_grads
		self.W_o_atprev -= self.W_o_atprev_grads

		self.w_i_ctprev -= self.w_i_ctprev_grads
		self.w_f_ctprev -= self.w_f_ctprev_grads
		self.w_o_c -= self.w_o_c_grads

		self.b_i -= self.b_i_grads
		self.b_f -= self.b_f_grads
		self.b_c -= self.b_c_grads
		self.b_o -= self.b_o_grads

	def get_Output_dim(self):
		return self.num_cel

	def initialize(self):
		self.layers = self.prev_recurrent + self.prev

		self.W_i_alprevs = [self.initializer.get((l.get_Output_dim(), self.get_Output_dim())) for l in self.layers]
		self.W_f_alprevs = [self.initializer.get((l.get_Output_dim(), self.get_Output_dim())) for l in self.layers]
		self.W_c_alprevs = [self.initializer.get((l.get_Output_dim(), self.get_Output_dim())) for l in self.layers]
		self.W_o_alprevs = [self.initializer.get((l.get_Output_dim(), self.get_Output_dim())) for l in self.layers]

		self.W_i_atprev = self.initializer.get((self.get_Output_dim(), self.get_Output_dim()))
		self.W_f_atprev = self.initializer.get((self.get_Output_dim(), self.get_Output_dim()))
		self.W_c_atprev = self.initializer.get((self.get_Output_dim(), self.get_Output_dim()))
		self.W_o_atprev = self.initializer.get((self.get_Output_dim(), self.get_Output_dim()))

		self.w_i_ctprev = self.initializer.get((1, self.get_Output_dim()))
		self.w_f_ctprev = self.initializer.get((1, self.get_Output_dim()))
		self.w_o_c = self.initializer.get((1, self.get_Output_dim()))

		self.b_i = self.initializer.get((1, self.get_Output_dim()))
		self.b_f = self.initializer.get((1, self.get_Output_dim()))
		self.b_c = self.initializer.get((1, self.get_Output_dim()))
		self.b_o = self.initializer.get((1, self.get_Output_dim()))

	def prop(self):
		len_prev_rec = len(self.prev_recurrent)

		self.i += [None]
		self.f += [None]
		self.tanh_zc += [None]
		self.c += [None]
		self.o += [None]
		self.a += [None]

		self.i[-1] = self.__sigm__(sum([
				np.dot(self.prev_recurrent[l_i].get_Output(t=0 if self.prev_recurrent[l_i].prop_idx >= self.prop_idx else -1), self.W_i_alprevs[l_i])
				for l_i in range(len(self.prev_recurrent))
			])\
			+ sum([
				np.dot(self.prev[l_i].get_Output(t=0), self.W_i_alprevs[len_prev_rec+l_i])
					for l_i in range(len(self.prev))
			])\
			+ np.dot(self.get_Output(t=-1), self.W_i_atprev)\
			+ self.w_i_ctprev * self.get_c(t=-1)\
			+ self.b_i
		)

		self.f[-1] = self.__sigm__(sum([
				np.dot(self.prev_recurrent[l_i].get_Output(t=0 if self.prev_recurrent[l_i].prop_idx >= self.prop_idx else -1), self.W_f_alprevs[l_i])
				for l_i in range(len(self.prev_recurrent))
			])\
			+ sum([
				np.dot(self.prev[l_i].get_Output(t=0), self.W_f_alprevs[len_prev_rec+l_i])
					for l_i in range(len(self.prev))
			])\
			+ np.dot(self.get_Output(t=-1), self.W_f_atprev)\
			+ self.w_f_ctprev * self.get_c(t=-1)\
			+ self.b_f
		)

		self.tanh_zc[-1] = self.__tanh__(
			sum([
				np.dot(self.prev_recurrent[l_i].get_Output(t=0 if self.prev_recurrent[l_i].prop_idx >= self.prop_idx else -1), self.W_c_alprevs[l_i])
				for l_i in range(len(self.prev_recurrent))
			])\
			+ sum([
				np.dot(self.prev[l_i].get_Output(t=0), self.W_c_alprevs[len_prev_rec+l_i])
					for l_i in range(len(self.prev))
			])\
			+ np.dot(self.get_Output(t=-1), self.W_c_atprev)\
			+ self.b_c
		)

		self.c[-1] = self.get_f(t=0) * self.get_c(t=-1) + self.get_i(t=0) * self.get_tanh_zc(t=0)

		self.o[-1] = self.__sigm__(sum([
				np.dot(self.prev_recurrent[l_i].get_Output(t=0 if self.prev_recurrent[l_i].prop_idx >= self.prop_idx else -1), self.W_o_alprevs[l_i])
				for l_i in range(len(self.prev_recurrent))
			])\
			+ sum([
				np.dot(self.prev[l_i].get_Output(t=0), self.W_o_alprevs[len_prev_rec+l_i])
					for l_i in range(len(self.prev))
			])\
			+ np.dot(self.get_Output(t=-1), self.W_o_atprev)\
			+ self.w_o_c * self.get_c(t=0)\
			+ self.b_o
		)

		self.a[-1] = self.get_o(t=0) * self.__tanh__(self.get_c(t=0))
		return self.get_Output(t=0)

	def reset(self, minibatch_size, init_cond = None):
		if init_cond is None:
			self.a = [np.zeros((minibatch_size, self.num_cel))]
			self.c = [np.zeros((minibatch_size, self.num_cel))]
		else:
			len_ = int(init_cond.shape[0]/2)
			self.a = [init_cond[:len_,:]]
			self.c = [init_cond[len_:,:]]
		self.tanh_zc = []
		self.i = []
		self.f = []
		self.o = []
		self.error_a = []
		self.error_o = []
		self.error_c = []
		self.error_f = []
		self.error_i = []

	def __save__dict__(self):
		return {
			"W_i_alprevs": self.W_i_alprevs,
			"W_f_alprevs": self.W_f_alprevs,
			"W_c_alprevs": self.W_c_alprevs,
			"W_o_alprevs": self.W_o_alprevs,

			"W_i_atprev": self.W_i_atprev,
			"W_f_atprev": self.W_f_atprev,
			"W_c_atprev": self.W_c_atprev,
			"W_o_atprev": self.W_o_atprev,

			"w_i_ctprev": self.w_i_ctprev,
			"w_f_ctprev": self.w_f_ctprev,
			"w_o_c": self.w_o_c,

			"b_i": self.b_i,
			"b_f": self.b_f,
			"b_c": self.b_c,
			"b_o": self.b_o
		}, [self.num_cel, self.name]

	def __load__dict__(self, d):
		self.W_i_alprevs = d["W_i_alprevs"]
		self.W_f_alprevs = d["W_f_alprevs"]
		self.W_c_alprevs = d["W_c_alprevs"]
		self.W_o_alprevs = d["W_o_alprevs"]

		self.W_i_atprev = d["W_i_atprev"]
		self.W_f_atprev = d["W_f_atprev"]
		self.W_c_atprev = d["W_c_atprev"]
		self.W_o_atprev = d["W_o_atprev"]

		self.w_i_ctprev = d["w_i_ctprev"]
		self.w_f_ctprev = d["w_f_ctprev"]
		self.w_o_c = d["w_o_c"]

		self.b_i = d["b_i"]
		self.b_f = d["b_f"]
		self.b_c = d["b_c"]
		self.b_o = d["b_o"]

class Convolution(Layer):
	def __init__(self, shape, kernel, step, nb_filters, name):
		super(Convolution, self).__init__(name)
		assert len(shape) == len(kernel)
		assert len(shape) == len(step)
		assert sum([shape[s]>=kernel[s] for s in range(len(shape))]) == len(shape)
		# assert sum(map(lambda x: x%2!=1, kernel))==0
		self.shape = tuple(shape)
		self.kernel_shape = (nb_filters,)+tuple(kernel)
		self.step = tuple(step)
		self.nb_filters = nb_filters
		self.output_dim = np.multiply.reduce((np.subtract(self.shape, self.kernel_shape[1:])//self.step)+1) * self.nb_filters

	def initialize(self):
		assert sum([l.get_Output_dim() for l in self.prev_recurrent + self.prev]) == np.multiply.reduce(self.shape)
		assert len(self.prev_recurrent + self.prev)==1 or len(self.prev_recurrent + self.prev) == self.shape[0]
		assert sum([(self.prev_recurrent + self.prev)[i].get_Output_dim() != (self.prev_recurrent + self.prev)[i+1].get_Output_dim() for i in range(len(self.prev_recurrent + self.prev)-1)]) == 0
		self.W = self.initializer.get(self.kernel_shape)
		self.b = self.initializer.get((self.nb_filters,))

	def get_error(self, t=0):
		return self.error[t-1].reshape((self.error[t-1].shape[0], self.nb_filters)+tuple((np.subtract(self.shape, self.kernel_shape[1:])//self.step)+1))

	def get_error_contribution(self, layer, t=0):

		if len(self.contributions) >= (-t)+1:
			aux = self.contributions[-t]
		else:
			error = self.get_error(t=t)
			error = np.moveaxis(error, 1, -1)
			shape_tmp = error.shape[:-1]
			error = error.reshape(-1, self.nb_filters)
			W_tmp = self.W.reshape(self.nb_filters, -1)

			error_x_M = np.dot(error, W_tmp)
			error_x_M = error_x_M.reshape(shape_tmp+self.kernel_shape[1:])
			aux = np.zeros((shape_tmp[0],) + tuple([shape_tmp[i]*self.step[i-1] for i in range(1,len(shape_tmp)) ]) + self.kernel_shape[1:])

			aux[tuple([slice(None)]+[slice(None,None,x) for x in self.step]+[slice(None) for x in self.step])] = error_x_M


			aux = np.lib.pad(aux,[(0,0)]+[(x-1,x-1) for x in self.kernel_shape[1:]]+[(0,0) for x in self.step] , 'constant', constant_values=0)

			shape = (aux.shape[0],) + self.shape + self.kernel_shape[1:]
			strides = list(aux.strides)
			num_dimensions = len(self.shape)
			for i in range(-1, -num_dimensions-1, -1):
				strides[i] -= strides[i-num_dimensions]

			aux = np.lib.stride_tricks.as_strided(aux, strides=strides)
			aux = aux[tuple([slice(None)]+[slice(x-1,None,None) for x in self.kernel_shape[1:]]+[slice(None) for x in self.kernel_shape[1:]])]

			aux = np.sum(aux, axis=tuple(range(-1, -num_dimensions-1, -1)))

			aux = aux.reshape(aux.shape[0], -1)

			self.contributions.insert(0, aux)

		idx = 0
		for l in self.prev_recurrent+self.prev:
			if l is not layer:
				idx+=l.get_Output_dim()
			else:
				return aux[:,idx:idx+layer.get_Output_dim()]

	def backprop_error(self,t=0):
		aux = 0
		if t != 0:
			for layer in self.next_recurrent:
				aux += layer.get_error_contribution(self, t=t+1)
		for layer in self.next:
			aux += layer.get_error_contribution(self, t=t)
		self.error = [aux]+self.error

	def reset_grads(self):
		self.b_grad = 0
		self.W_grad = 0

	def compute_gradients(self):
		num_t = len(self.error)
		num_dimensions = len(self.shape)

		self.b_grad = 0
		self.W_grad = 0

		for t in range(0,-num_t, -1):
			error = self.get_error(t=t)
			self.b_grad += np.sum(error, axis=tuple([0]+list(range(2, num_dimensions+2))))
			error = np.moveaxis(error, 1, -1)
			shape_tmp = error.shape[:-1]
			error = error.reshape(-1, self.nb_filters)

			self.W_grad += np.dot(error.T, self.x_M[t-1]).reshape(self.W.shape)

	def apply_to_gradients(self, func):
		self.b_grad = func(self.b_grad)
		self.W_grad = func(self.W_grad)

	def get_params(self):
		return [self.b, self.W]

	def update(self):
		self.__update_b__()
		self.__update_W__()

	def __update_W__(self):
		self.W -= self.W_grad

	def __update_b__(self):
		self.b -= self.b_grad

	def prop(self):
		inp = np.concatenate(
			[l.get_Output(t=0 if l.prop_idx >= self.prop_idx else -1) for l in self.prev_recurrent]+
			[l.get_Output(t=0) for l in self.prev], axis=1)
		shape = (inp.shape[0],) + self.kernel_shape[1:] + tuple((np.subtract(self.shape, self.kernel_shape[1:])//self.step)+1)
		strides = [inp.strides[-1]]
		for sh in reversed(self.shape):
			strides.insert(0, strides[0]*sh)
		strides = (strides[0],) + tuple(strides[1:]) + tuple(np.multiply(strides[1:], self.step))
		x_M = np.lib.stride_tricks.as_strided(inp, shape=shape, strides=strides)
		x_M = np.moveaxis(x_M, list(range(1, len(self.shape)+1)), list(range(-len(self.shape),0)))
		new_shape = x_M.shape
		x_M = x_M.reshape((np.multiply.reduce(new_shape[:len(self.shape)+1]), np.multiply.reduce(new_shape[len(self.shape)+1:]) ))

		self.x_M.append(x_M)
		
		W_tmp = self.W.reshape(self.nb_filters, -1).T

		out = (np.dot(x_M, W_tmp) + self.b)

		out = out.reshape(new_shape[:len(self.shape)+1]+ (-1,))
		out = np.moveaxis(out, -1, 1)

		out = out.reshape(-1, self.output_dim)

		self.a = self.a + [out]
		return out

	def reset(self, minibatch_size):
		if len(self.next_recurrent)>0:
			self.a = [np.zeros((minibatch_size, self.get_Output_dim()))]
		else:
			self.a = []
		self.error = []
		self.x_M = []
		self.contributions = []

	def __save__dict__(self):
		return {
			'W': self.W,
			'b': self.b
		}, [self.shape, self.kernel_shape[1:], self.step, self.nb_filters, self.name]

	def __load__dict__(self, d):
		self.W = d['W']
		self.b = d['b']

class Activation(Layer):
	def __init__(self, activation_function, name):
		super(Activation, self).__init__(name)
		self.set_activation_function(activation_function)

	def set_activation_function(self, activation_function):
		self.act_fun = activation_function

		if activation_function == "sigmoid":
			self.ff = lambda z: (np.tanh(z/2)+1)/2
			self.derivative_ff = lambda a: a*(1-a)
			self.multiplication = lambda a,b: a*b

		elif activation_function == "tanh":
			self.ff = lambda z: np.tanh(z)
			self.derivative_ff = lambda a: 1 - a**2
			self.multiplication = lambda a,b: a*b

		elif activation_function == "dummy_softmax":
			def ff_ (z):
				exp = np.exp(z)
				return exp/np.sum(exp, axis=1)[:,np.newaxis]
			self.ff = ff_
			self.derivative_ff = lambda a: a*(1-a)
			self.multiplication = lambda a,b: a*b

		elif activation_function == "softmax":
			def ff(z):
				exp = np.exp(z)
				return exp/np.sum(exp, axis=1)[:,np.newaxis]
			self.ff = ff
			def softmax_calc(x):
				x=x[np.newaxis]
				a = -np.dot(x.T,x)
				np.fill_diagonal(a, x*(1-x))
				return a
			self.softmax_calc = softmax_calc
			def derivative_ff(a):
				b = np.apply_along_axis(self.softmax_calc, 1, a)
				return b
			self.derivative_ff = derivative_ff
			self.multiplication = lambda a,b: np.einsum('ij,ijk->ik', a, b)

		elif activation_function == "relu":
			self.ff = lambda z: np.maximum(z, 0)
			self.derivative_ff = lambda a: np.sign(a)
			self.multiplication = lambda a,b: a*b

		elif activation_function == "leaky_relu":
			self.ff = lambda z: np.maximum(z, 0)
			self.derivative_ff = lambda a: np.maximum(np.sign(a), 0.01)
			self.multiplication = lambda a,b: a*b

		elif activation_function == "linear":
			self.ff = lambda z: z
			self.derivative_ff = lambda a: np.full(a.shape, 1)
			self.multiplication = lambda a,b: a*b
		else:
			raise ValueError("Activation function not defined")

	def get_Output_dim(self):
		return sum([l.get_Output_dim() for l in (self.prev_recurrent+self.prev)])

	def get_data_of_layer(self, x, layer):
		i = (self.prev_recurrent+self.prev).index(layer)
		init = sum([l.get_Output_dim() for l in (self.prev_recurrent+self.prev)[:i]])
		end = init+layer.get_Output_dim()
		return x[:,init:end]

	def get_error_contribution(self, layer, t=0):
		return self.get_data_of_layer(self.multiplication(
					self.get_error(t=t),
					self.derivative_ff(self.get_Output(t=t))
				), layer)

	def backprop_error(self,t=0):
		#  this function will be used during the BP. to calculate the error
		aux = 0
		if t != 0:
			for layer in self.next_recurrent:
				aux += layer.get_error_contribution(self, t=t+1)
		for layer in self.next:
			aux += layer.get_error_contribution(self, t=t)
		self.error = [aux]+self.error

	def get_params(self):
		return []

	def prop(self):
		out = self.ff(np.concatenate([l.get_Output() for l in self.prev_recurrent + self.prev], axis=0))
		self.a = self.a + [out]
		return out

	def __save__dict__(self):
		return {}, [self.act_fun, self.name]

class Loss(Layer):
	def __init__(self, loss, name):
		super(Loss, self).__init__(name)
		self.loss = loss
		self.output_dim = sum([l.get_Output_dim() for l in self.prev_recurrent + self.prev])

	def ff(self, a, y):
		if self.loss == "mse":
			return sum((y-a)**2)/2

		elif self.loss == "ce1":
			return -(y*np.log(a+0.000001))-((1-y)*np.log(1.000001-a))

		elif self.loss == "ce2":
			return -(y*np.log(a+0.000001))

		elif self.loss == "external":
			return a

	def grad(self, a, y):
		# returns the gradient with respect to all the components of a.
		# Thus the dimension of the output np.array is the same that a, i.e. [output_dim, 1]
		if self.loss == "mse":
			return a - y

		elif self.loss == "ce1":
			if np.sum(a == 0) == 0 and np.sum(a==1) == 0:
				# if np.isnan(-y/(a) + (1-y)/(1-a)).any():
				# 	pass
				return -(y/(a)) + (1-y)/(1-a)
			else:
				return -(y/(a+0.000001)) + (1-y)/(1.000001-a)

		elif self.loss == "ce2":
			if np.sum(a == 0) == 0:
				# if np.isnan(-y/(a) + (1-y)/(1-a)).any():
				# 	pass
				return -(y/(a))
			else:
				return -(y/(a+0.000001))

		elif self.loss == "external":
			return y

	def addNext(self, layer):
		assert False, "Loss layer do not have next layers"

	def addNextRecurrent(self, layer):
		assert False, "Loss layer do not have next layers"

	def __addPrev__(self, layer):
		self.prev.append(layer)
		assert len(self.prev) == 1

	def __addPrevRecurrent__(self, layer):
		assert False, "Loss layer do not have recurrent layer prev"

	def get_Output(self, t=0):
		return self.a[t-1]

	def get_error(self, t=0):
		assert False, "no error in loss layers"

	def get_error_contribution(self, layer, t=0):
		if self.desired_output is not None:
			return self.grad(self.prev[0].get_Output(t=t), self.desired_output)
		else:
			return np.zeros(self.prev[0].get_Output(t=t).shape)
	def backprop_error(self,t=0, desired_output=None):
		self.desired_output = desired_output

	def get_params(self):
		return []

	def prop(self, desired_output=None):
		out = self.prev[0].get_Output()
		if desired_output is not None:
			out = self.ff(out, desired_output)
		self.a = self.a + [out]
		return out

	def __save__dict__(self):
		return {}, [self.loss, self.name]

class Dropout(Layer):
	def __init__(self, probability, name):
		super(Dropout, self).__init__(name)
		self.probability = probability
		self.output_dim = None

	def get_Output_dim(self):
		if self.output_dim == None:
			self.output_dim = sum([l.get_Output_dim() for l in (self.prev_recurrent+self.prev)])
		return self.output_dim

	def get_data_of_layer(self, x, layer):
		i = (self.prev_recurrent+self.prev).index(layer)
		init = sum([l.get_Output_dim() for l in (self.prev_recurrent+self.prev)[:i]])
		end = init+layer.get_Output_dim()
		return x[:,init:end]

	def get_error_contribution(self, layer, t=0):
		return self.get_data_of_layer(self.get_error(t=t),layer)*self.get_data_of_layer(self.get_Output(t=t),layer)

	def backprop_error(self,t=0):
		#  this function will be used during the BP. to calculate the error
		aux = 0
		if t != 0:
			for layer in self.next_recurrent:
				aux += layer.get_error_contribution(self, t=t+1)
		for layer in self.next:
			aux += layer.get_error_contribution(self, t=t)
		aux = aux * self.mask
		self.error = [aux]+self.error

	def reset(self, minibatch_size):
		if len(self.next_recurrent)>0:
			self.a = [np.zeros((minibatch_size, self.get_Output_dim()))]
		else:
			self.a = []
		self.mask = np.random.rand(minibatch_size, self.get_Output_dim()) > self.probability
		self.error = []

	def get_params(self):
		return []

	def prop(self, desired_output=None):
		out = np.concatenate([l.get_Output(t=0 if l.prop_idx >= self.prop_idx else -1) for l in  self.prev_recurrent]+[l.get_Output(t=0) for l in self.prev], axis = 1) * self.mask
		self.a = self.a + [out]
		return out

	def __save__dict__(self):
		return {}, [self.probability, self.name]

class MaxPooling(Layer):
	def __init__(self, shape, kernel, name):
		super(MaxPooling, self).__init__(name)
		assert len(shape) == len(kernel)
		assert sum([shape[s]>=kernel[s] for s in range(len(shape))]) == len(shape)
		self.shape = tuple(shape)
		self.kernel_shape = tuple(kernel)
		self.output_dim = np.multiply.reduce((np.subtract(self.shape, self.kernel_shape)//self.kernel_shape)+1)

	def get_Output_dim(self):
		return self.output_dim

	def get_data_of_layer(self, x, layer):
		i = (self.prev_recurrent+self.prev).index(layer)
		init = sum([l.get_Output_dim() for l in (self.prev_recurrent+self.prev)[:i]])
		end = init+layer.get_Output_dim()
		return x[:,init:end]

	def get_error_contribution(self, layer, t=0):
		return self.get_data_of_layer(self.get_error(t=t),layer)#*self.get_data_of_layer(self.get_Output(t=t),layer)

	def __reshape_idex__(self, k):
		i = np.arange(0,self.shape[k]-self.kernel_shape[k]+1,self.kernel_shape[k])
		i = np.expand_dims(i, axis=0)
		for d in range(len(self.kernel_shape)):
			if d < k:
				i = np.expand_dims(i, axis=0)
			elif d > k:
				i = np.expand_dims(i, axis=-1)
		return i

	def backprop_error(self,t=0):
		#  this function will be used during the BP. to calculate the error
		aux = 0
		if t != 0:
			for layer in self.next_recurrent:
				aux += layer.get_error_contribution(self, t=t+1)
		for layer in self.next:
			aux += layer.get_error_contribution(self, t=t)
		# aux = aux * self.mask

		d = np.zeros((self.get_Output(t=t).shape[0],)+self.shape)

		i = []
		for k in range(len(self.kernel_shape)):
			i.append((self.masks[t-1][k]+self.__reshape_idex__(k)).flatten())

		if aux is not 0:
			i.insert(0, np.repeat(np.arange(aux.shape[0]), i[0].shape[0]//aux.shape[0]))
			d[tuple(i)] = aux.flatten()

		aux = d.reshape(d.shape[0], np.multiply.reduce(d.shape[1:]))

		self.error = [aux]+self.error

	def reset(self, minibatch_size):
		if len(self.next_recurrent)>0:
			self.a = [np.zeros((minibatch_size, self.get_Output_dim()))]
		else:
			self.a = []
		self.masks = []
		self.error = []

	def get_params(self):
		return []

	def initialize(self):
		assert sum([l.get_Output_dim() for l in self.prev_recurrent + self.prev]) == np.multiply.reduce(self.shape), "%s: %s --> %d" % (self.name, self.shape, sum([l.get_Output_dim() for l in self.prev_recurrent + self.prev]))
		assert sum([(self.prev_recurrent + self.prev)[i].get_Output_dim() != (self.prev_recurrent + self.prev)[i+1].get_Output_dim() for i in range(len(self.prev_recurrent + self.prev)-1)]) == 0

	def prop(self, desired_output=None):
		out = np.concatenate([l.get_Output(t=0 if l.prop_idx >= self.prop_idx else -1) for l in self.prev_recurrent]+[l.get_Output(t=0) for l in self.prev], axis = 1)

		shape_ = (out.shape[0],) + tuple((np.subtract(self.shape, self.kernel_shape)//self.kernel_shape)+1) + self.kernel_shape
		strides = [out.strides[-1]]
		for sh in reversed(self.shape):
			strides.insert(0, strides[0]*sh)
		strides = (strides[0],) + tuple(np.multiply(strides[1:], self.kernel_shape)) + tuple(strides[1:])
		a = np.lib.stride_tricks.as_strided(out, shape=shape_, strides=strides)
		a = a.reshape(a.shape[:-len(self.kernel_shape)]+(np.multiply.reduce(a.shape[-len(self.kernel_shape):]),))
		self.masks.append(
			np.unravel_index(np.argmax(a, axis = len(a.shape)-1), self.kernel_shape)
		)

		out = np.amax(a, axis = len(a.shape)-1)
		out = out.reshape(out.shape[0], np.multiply.reduce(out.shape[1:]))

		self.a = self.a + [out]
		return out

	def __save__dict__(self):
		return {}, [self.shape, self.kernel_shape, self.name]



from setuptools import setup, find_packages

setup(name='wbnn',
      version='1.2.12',
      description='White Box Neural Networks. Neural Network and Deep Learning toolkit developed for educational purposes.',
      url='https://github.com/develask/White-Box-Neural-Networks',
      project_urls={
            "Documentation": "https://github.com/develask/White-Box-Neural-Networks/wiki",
            "Source Code": "https://github.com/develask/White-Box-Neural-Networks",
      },
      author='Mikel de Velasco Vázquez',
      author_email='develascomikel@gmail.com',
      license='MIT',
      packages=['wbnn', 'wbnn.visualLogs'],
      entry_points={
          'console_scripts': [
              'wbnnplotscmd = wbnn.visualLogs.plot:main',
              'wbnnplots = wbnn.visualLogs.gui:main'
          ]
      },
      install_requires=['numpy', 'matplotlib'],
      zip_safe=False)


# python setup.py sdist
# python setup.py sdist upload



import time, os
import numpy as np
import json
from . import layers, optimizers

class NN():
	def __init__(self, name):
		self.name = name
		self.inputs = []
		self.output_layers = []
		self.info = {} 

	def calculate_layer_order(self):
		"""
		Since NN object contains a graph representation,
		this function calculates the propagation order.
		Using the reverse order, work for backpropagation.
		"""
		self.prop_order = self.inputs.copy()
		i = 0
		while i < len(self.prop_order):
			layer = self.prop_order[i]
			for l in layer.next:
				if sum(map(lambda l_: not l_ in self.prop_order and not l_ is l, l.prev)) == 0 and not l in self.prop_order:
					self.prop_order += [l]
			i+=1

		for l_i in range(len(self.prop_order)):
			self.prop_order[l_i].prop_idx = l_i


	def initialize(self, init_layers = True):
		"""Initialize each layer, and generate the layer order."""
		if not hasattr(self, 'prop_order'):
			self.calculate_layer_order()
		for layer in self.prop_order:
			if init_layers:
				layer.initialize()
			if isinstance(layer, layers.Loss):
				self.output_layers.append(layer)

	def add_inputs(self, *layers):
		for layer in layers:
			self.inputs.append(layer)

	def get_Output(self, t=0):
		return [l.get_Output(t=t) for l in self.output_layers]

	def prop(self, inp, desired_output=None):
		for layer in self.prop_order:
			layer.reset(inp[0][0].shape[0])
		out = []
		out_t_cop = None
		for t in range(len(inp)):
			inp_t_cop = inp[t].copy()
			if desired_output and t - (len(inp)-len(desired_output)) >= 0:
				out_t_cop = desired_output[t - (len(inp)-len(desired_output))].copy()
			for layer in self.prop_order:
				if isinstance(layer, layers.Input) or (isinstance(layer, layers.Loss) and out_t_cop):
					if isinstance(layer, layers.Input):
						layer.prop(inp_t_cop.pop(0))
					else:
						layer.prop(out_t_cop.pop(0))
				else:
					layer.prop()

			out.append(self.get_Output())
		return out

	def backprop(self, inp, desired_output):
		out = self.prop(inp.copy(), desired_output)

		for t in range(0,-len(inp),-1):
			for layer in self.prop_order[-1:len(self.inputs)-1:-1]:
				if isinstance(layer, layers.Loss):
					if t <= -len(desired_output):
						layer.backprop_error(t=t, desired_output=None)
					else:
						layer.backprop_error(t=t, desired_output=desired_output[t-1][self.output_layers.index(layer)])
				else:
					layer.backprop_error(t=t)

		for layer in self.prop_order[:len(self.inputs)-1:-1]:
			layer.compute_gradients()
		return out

	def reset_gradients(self):
		for k in range(len(self.inputs),len(self.prop_order)):
			self.prop_order[k].reset_grads()

	def apply_to_gradients(self, func):
		for k in range(len(self.inputs),len(self.prop_order)):
			self.prop_order[k].apply_to_gradients(func)

	def get_params(self):
		params = []
		for l in self.prop_order:
			params += l.get_params()
		return params

	def get_gradients(self):
		gradients = []
		def insert_on_grads(grad):
			gradients.append(grad)
			return grad
		self.apply_to_gradients(insert_on_grads)
		return gradients

	def set_gradients(self, gradients):
		self.apply_to_gradients(lambda x: gradients.pop(0))

	def compute_minibatch_grad(self, mini_batch):
		self.reset_gradients()
		out = self.backprop(*mini_batch)
		loss = sum([sum([ np.sum(out_o) for out_o in out_t]) for out_t in out[len(mini_batch[0])-len(mini_batch[1]):]])

		# Normalizar los gradientes
		self.apply_to_gradients(lambda grad: grad/mini_batch[1][0][0].shape[0])
		return self.get_gradients(),loss

	def update_model(self):
		for k in range(len(self.inputs),len(self.prop_order)):
			self.prop_order[k].update()

	def get_loss_of_data(self, data):
		len_tr = data[1][0][0].shape[0]
		out = self.prop(data[0].copy(), data[1].copy())
		loss = sum([sum([ np.sum(out_o) for out_o in out_t]) for out_t in out[len(data[0])-len(data[1]):]])
		return loss/len_tr

	def save(self, name = None):
		dir_ = time.strftime(name if name is not None else self.name + "_%Y-%m-%d_%H-%M")
		try:
			os.makedirs(dir_)
		except Exception as e:
			dir_ = dir_+time.strftime("_%Y-%m-%d_%H-%M")
			os.makedirs(dir_)
		inf = {
			'prop_order': [id(l) for l in self.prop_order],
			'info': self.info
		}
		for layer in self.prop_order:
			l_id = id(layer)
			inf[str(l_id)] = {
				'next': [id(l) for l in layer.next],
				'prev': [id(l) for l in layer.prev],
				'next_recurrent': [id(l) for l in layer.next_recurrent],
				'prev_recurrent': [id(l) for l in layer.prev_recurrent],
			}
			layer.save(dir_)
		file_ = json.dumps(inf)
		with open(dir_+"/setup.json", "w") as f:
			f.write(file_)

	def load(dir_):
		with open(dir_+"/setup.json", "r") as f:
			inf = json.loads(f.read())
			layers_ = [layers.Layer.load(dir_+'/'+str(id_)) for id_ in inf['prop_order']]
			layers_id = [str(x) for x in inf['prop_order']]
			external_info = inf['info']
			del inf['info']
			del inf['prop_order']
			for l in inf:
				idx = layers_id.index(str(l))
				layers_[idx].next = [layers_[layers_id.index(str(ln))] for ln in inf[l]['next']]
				layers_[idx].prev = [layers_[layers_id.index(str(ln))] for ln in inf[l]['prev']]
				layers_[idx].next_recurrent = [layers_[layers_id.index(str(ln))] for ln in inf[l]['next_recurrent']]
				layers_[idx].prev_recurrent = [layers_[layers_id.index(str(ln))] for ln in inf[l]['prev_recurrent']]

			inps = list(filter(lambda x: isinstance(x, layers.Input), layers_))
			nn = NN(dir_)
			nn.info = external_info
			for inp in inps:
				nn.add_inputs(inp)

			nn.initialize(init_layers = False)
			return nn



import numpy as np

class Initializer():
	def get(self, shape):
		raise NotImplementedError( "Should have implemented this" )

class Constant(Initializer):
	def __init__(self, constant):
		self.constant = constant

	def get(self, shape):
		return np.full(shape, self.constant)


class RandomNormal(Initializer):
	def __init__(self, mean=0.0, stddev=None):
		self.mean = mean
		self.stddev = stddev

	def get(self, shape):
		stddev = self.stddev
		if stddev == None:	# Xavier initialization
			stddev = np.sqrt(2/sum(shape))
		return np.random.normal(loc=self.mean, scale=stddev, size=shape)

class RandomUniform(Initializer):
	def __init__(self, minVal=-0.01, maxVal=0.01):
		self.minVal = minVal
		self.maxVal = maxVal
	
	def get(self, shape):
		return np.random.uniform(low=self.minVal, high=self.maxVal, size=shape)
		
		



import matplotlib
matplotlib.use('TkAgg')
import json
import sys
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
from matplotlib.collections import PolyCollection
from matplotlib.widgets import Slider, RadioButtons
import matplotlib.ticker as ticker
from matplotlib import colors
import numpy as np
import time
import math

COLORS = [
	{
			'name': 'Deep Orange',
			'color': '#ff6f42',
			'active': '#ca4a06',
			'disabled': '#f2cbba'
		},
		{
			'name': 'Google Blue',
			'color': '#4184f3',
			'active': '#3a53c5',
			'disabled': '#cad8fc'
		},
		{
			'name': 'Google Yellow',
			'color': '#f4b400',
			'active': '#db9200',
			'disabled': '#f7e8b0'
		},
		{
			'name': 'Google Green',
			'color': '#0f9d58',
			'active': '#488046',
			'disabled': '#c2e1cc'
		},
		{
			'name': 'Pink',
			'color': '#ef6191',
			'active': '#ca1c60',
			'disabled': '#e9b9ce'
		},
		{
			'name': 'Purple',
			'color': '#aa46bb',
			'active': '#5c1398',
			'disabled': '#d7bce6'
		},
		{
			'name': 'Teal',
			'color': '#00abc0',
			'active': '#47828e',
			'disabled': '#c2eaf2'
		},
		{
			'name': 'Lime',
			'color': '#9d9c23',
			'active': '#7f771d',
			'disabled': '#f1f4c2'
		},
	{
			'name': 'Google Red',
			'color': '#db4437',
			'active': '#8f2a0c',
			'disabled': '#e8c6c1'
		},
		{
			'name': 'Indigo',
			'color': '#5b6abf',
			'active': '#3e47a9',
			'disabled': '#c5c8e8'
		},
		{
			'name': 'Deep Teal',
			'color': '#00786a',
			'active': '#2b4f43',
			'disabled': '#bededa'
		},
		{
			'name': 'Deep Pink',
			'color': '#c1175a',
			'active': '#75084f',
			'disabled': '#de8cae'
		},
		{
			'name': 'Gray',
			'color': '#9E9E9E',
			'active': '#424242',
			'disabled': '#F5F5F5'
		}
]

def to_rgb(c):
	r = int(c[1:3],16)
	g = int(c[3:5],16)
	b = int(c[5:7],16)
	return r,g,b


class Plotter():
	def __init__(self, files_to_read):
		self.files_to_read = files_to_read

		show_folder =  len(set(['/'.join(file_to_read.split('/')[:-1]) for file_to_read in files_to_read])) > 1

		common_name = [file_to_read.split('/') for file_to_read in files_to_read]
		from_idx = 0
		for i in range(len(common_name[0])):
			w = common_name[0][i]
			for ws2 in common_name[1:]:
				if ws2[i] != w:
					from_idx = i
					break
			if from_idx != 0:
				break
		if from_idx == 0:
			from_idx = len(common_name[0])

		self.logs = {}
		for file_to_read in files_to_read:
			format_ = file_to_read.split('/')[-1].split('_')[0]
			if format_ != self.type:
				continue

			name = '_'.join(file_to_read.split('/')[-1].split('_')[1:])[:-4]
			if show_folder:
				name = name + ' ('+'/'.join(file_to_read.split('/')[from_idx:-1])+')'

			my_logs = self.read_file(file_to_read)
			self.logs[name] = my_logs


		if len(list(self.logs)) > 0:
			self.plot(self.logs)

	def read_file(self, file):
		return None

	def plot(self):
		print("No plot function")

class Histogram(Plotter):
	def __init__(self, files_to_read):
		self.type = 'hist'
		super(Histogram, self).__init__(files_to_read)

	def read_file(self, file_to_read):
		with open(file_to_read) as f:
			logs = f.read().splitlines()
		return [ ( float(x.split('\t')[1]), json.loads(x.split('\t')[2].replace('(','[').replace(')',']')) ) for x in logs]

	def plot(self, logs_dict):

		num_cols = math.ceil(math.sqrt(len(list(logs_dict))))
		if len(list(logs_dict)) > num_cols*(num_cols-1):
			num_rows = num_cols
		else:
			num_rows = num_cols-1

		# fig = plt.figure()
		fig = plt.figure(figsize=(num_cols*4, num_rows*4), dpi=80)
		num = 1
		self.axes = []
		for name in logs_dict:
			logs = logs_dict[name]
			color = COLORS[(num-1) % len(COLORS)]
			def polygon_under_graph(xlist, ylist):
					return [(xlist[0], 0.)] + list(zip(xlist, ylist)) + [(xlist[-1], 0.)]

			def color_(i):
				back = to_rgb(color['active'])
				front = to_rgb(color['disabled'])
				porcent = i / len(logs)
				def take(min_, max_, porcent):
					return ( min_ + ((max_-min_)*porcent) ) / 255
				return take(back[0], front[0], porcent), take(back[1], front[1], porcent), take(back[2], front[2], porcent), 0.8


			ax = fig.add_subplot(num_rows, num_cols, num, projection='3d')
			self.axes.append(ax)

			temps = []
			zs = range(len(logs))

			my_mod = int(len(logs)/10)
			if my_mod == 0:
				my_mod = 1

			#filer
			logs, zs = zip(*[ (x,i) for x,i in zip(logs, zs) if i%my_mod==0])

			val_max = 0
			num_vals = np.sum(np.asarray([x[1] for x in logs])[:,0,1])
			left_lim = 0
			right_lim = 0

			for (width, bar), z in zip(logs, zs):
				xs = [x[0]+width/2 for x in bar]
				ys = [x[1]/(width*num_vals) for x in bar]
				if max(ys) > val_max:
					val_max = max(ys)
				if min(xs) < right_lim:
					right_lim = min(xs)
				if max(xs) > left_lim:
					left_lim = max(xs)

				temps.append(polygon_under_graph(xs, ys))

				# ax.plot(xs, ys, z, zdir='y', color=(1,1,1), alpha=1, linewidth=1)

			poly = PolyCollection(temps, facecolors=[color_(x) for x in range(len(logs))], edgecolors=(1,1,1,0.8))
			ax.add_collection3d(poly, zs=zs, zdir='y')

			ax.set_xlabel('value')
			ax.set_ylabel('mini-batch')
			ax.set_xlim(right_lim, left_lim)
			ax.set_ylim(zs[-1], zs[0])
			ax.set_zlim(0, val_max)
			plt.title(name)
			num += 1

		def update(val):
			num =1
			new_axes = []
			for ax, name in zip(self.axes, list(logs_dict)):
				logs = logs_dict[name]
				color = COLORS[(num -1) % len(COLORS)]
				ax.remove()
				ax = fig.add_subplot(num_rows, num_cols, num, projection='3d')
				new_axes.append(ax)
				temps = []
				zs = range(len(logs))

				my_mod = int(len(logs)/int(val))
				if my_mod == 0:
					my_mod = 1

				#filer
				logs, zs = zip(*[ (x,i) for x,i in zip(logs, zs) if i%(my_mod)==0])

				def color_(i):
					back = to_rgb(color['active'])
					front = to_rgb(color['disabled'])
					porcent = i / len(logs)
					def take(min_, max_, porcent):
						return ( min_ + ((max_-min_)*porcent) ) / 255
					return take(back[0], front[0], porcent), take(back[1], front[1], porcent), take(back[2], front[2], porcent), 0.8

				val_max = 0
				num_vals = np.sum(np.asarray([x[1] for x in logs])[:,0,1])
				left_lim = 0
				right_lim = 0

				for (width, bar), z in zip(logs, zs):
					xs = [x[0]+width/2 for x in bar]
					ys = [x[1]/(width*num_vals) for x in bar]
					if max(ys) > val_max:
						val_max = max(ys)
					if min(xs) < right_lim:
						right_lim = min(xs)
					if max(xs) > left_lim:
						left_lim = max(xs)
					temps.append(polygon_under_graph(xs, ys))

				poly = PolyCollection(temps, facecolors=[color_(x) for x in range(len(logs))], edgecolors=(1,1,1,0.8))
				ax.add_collection3d(poly, zs=zs, zdir='y')

				ax.set_xlabel('value')
				ax.set_ylabel('mini-batch')
				ax.set_xlim(right_lim, left_lim)
				ax.set_ylim(zs[-1], zs[0])
				ax.set_zlim(0, val_max)
				plt.title(name)
				num+=1
			self.axes = new_axes

		fig.subplots_adjust(top=0.99,bottom=0.1, left=0.01, right=0.99)

		axcolor = 'lightgoldenrodyellow'
		axsmooth = plt.axes([0.25, 0.025, 0.60, 0.03], facecolor=axcolor)
		samp = Slider(axsmooth, 'samples', 5, 50, valinit=10, valstep=1)
		samp.on_changed(update)

		plt.show()

class Scalar(Plotter):
	def __init__(self, files_to_read):
		self.type = 'scalar'
		super(Scalar, self).__init__(files_to_read)

	def read_file(self, file_to_read):
		with open(file_to_read) as f:
			logs = f.read().splitlines()
		return [ (float(x.split('\t')[0]), float(x.split('\t')[1])) for x in logs]


	def plot(self, logs_dict, smooth_constant=1):
												# ['relative-time', 'real-time', 'iteration']

		names = list(logs_dict)
		names = sorted(names, key=lambda x: len(logs_dict[x]), reverse=True)
		mode = 'relative-time'
		fig, self.ax = plt.subplots()
		if mode == 'relative-time':
			time_min = min([logs_dict[name][0][0] for name in logs_dict])

			formatter = matplotlib.ticker.FuncFormatter(lambda ms, x: time.strftime('%H:%M:%S', time.gmtime(ms)))
			self.ax.xaxis.set_major_formatter(formatter)

		if mode == 'real-time':
			formatter = matplotlib.ticker.FuncFormatter(lambda ms, x: time.strftime("%b %d %H:%M:%S", time.gmtime(ms)))
			self.ax.xaxis.set_major_formatter(formatter)

		def mean(list_):
			return sum(list_)/len(list_)

		def to_y(x, max_x = 700, max_y = 30):
			if x > max_x:
				x = max_x
			return max_y/(max_x**2) * x**2 - max_y/(max_x**2)

		def smoothed_list(values, smooth_constant):
			smooth_dist = int(smooth_constant * to_y(len(values)))
			return [ mean(values[(i-smooth_dist if i-smooth_dist>=0 else 0):i+smooth_dist+1]) for i in range(len(values))]

		def update(val):
			mode = radio.value_selected
			if mode == 'relative-time':
				time_min = min([logs_dict[name][0][0] for name in names])
				formatter = matplotlib.ticker.FuncFormatter(lambda ms, x: time.strftime('%H:%M:%S', time.gmtime(ms)))
				self.ax.xaxis.set_major_formatter(formatter)

			if mode == 'real-time':
				formatter = matplotlib.ticker.FuncFormatter(lambda ms, x: time.strftime("%b %d %H:%M:%S", time.gmtime(ms)))
				self.ax.xaxis.set_major_formatter(formatter)

			if mode == 'iteration':
				self.ax.xaxis.set_major_formatter(ticker.ScalarFormatter(useMathText=True))


			i = 0
			for name in names:
				logs = logs_dict[name]
				times, values = zip(*logs)
				time_min = times[0]
				if mode == 'relative-time':
					times = [x-time_min for x in times]
				if mode == 'iteration':
					times = list(range(len(times)))
				smooth_values = smoothed_list(values, samp.val)
				backs[i].set_ydata(values)
				backs[i].set_xdata(times)
				i+=1

			i = 0
			for name in names:
				logs = logs_dict[name]
				times, values = zip(*logs)
				time_min = times[0]
				if mode == 'relative-time':
					times = [x-time_min for x in times]
				if mode == 'iteration':
					times = list(range(len(times)))
				smooth_values = smoothed_list(values, samp.val)
				legends[i].set_ydata(smooth_values)
				legends[i].set_xdata(times)
				i+=1

			self.ax.relim()
			self.ax.autoscale_view()

		legends = []
		backs = []
		c = 0
		for name in names:
			logs = logs_dict[name]
			times, values = zip(*logs)
			time_min = times[0]
			if mode == 'relative-time':
				times = [x-time_min for x in times]
			if mode == 'iteration':
				times = list(range(len(times)))
			color = COLORS[c % len(COLORS)]
			smooth_values = smoothed_list(values, smooth_constant)
			b, = plt.plot(times,values, '-', color=color['disabled'])
			backs.append(b)
			c+=1
		c = 0
		for name in names:
			logs = logs_dict[name]
			times, values = zip(*logs)
			time_min = times[0]
			if mode == 'relative-time':
				times = [x-time_min for x in times]
			if mode == 'iteration':
				times = list(range(len(times)))
			color = COLORS[c % len(COLORS)]
			smooth_values = smoothed_list(values, smooth_constant)
			leg, = plt.plot(times,smooth_values, '-', color=color['active'], label=name)
			legends.append(leg)
			c+=1

		# plt.title("Scalar over time")

		plt.legend(handles=legends)

		plt.gcf().autofmt_xdate()
		plt.grid()
		fig.subplots_adjust(bottom=0.25)

		axcolor = 'lightgoldenrodyellow'
		axsmooth = plt.axes([0.35, 0.025, 0.5, 0.03], facecolor=axcolor)
		samp = Slider(axsmooth, 'Smooth', 0.01, 2.0, valinit=1.0)
		samp.on_changed(update)

		rax = plt.axes([0.025, 0.025, 0.2, 0.1], facecolor=axcolor)
		radio = RadioButtons(rax, ['relative-time', 'real-time', 'iteration'], active=0)

		radio.on_clicked(update)

		plt.show()

class Image(Plotter):
	def __init__(self, files_to_read):
		self.type = 'image'
		super(Image, self).__init__(files_to_read)

	def read_file(self, file_to_read):
		data = np.load(file_to_read)
		return data['times'], data['values']

	def plot(self, logs_dict):
		num_cols = math.ceil(math.sqrt(len(list(logs_dict))))
		if len(list(logs_dict)) > num_cols*(num_cols-1):
			num_rows = num_cols
		else:
			num_rows = num_cols-1

		names = list(logs_dict)

		_min = 0
		_max = 0
		for title in names:
			new_min = np.min(logs_dict[title][1])
			new_max = np.max(logs_dict[title][1])
			if new_min < _min:
				_min = new_min
			if new_max > _max:
				_max = new_max

		# fig = plt.figure()
		fig, axeslist = plt.subplots(ncols=num_cols, nrows=num_rows)
		if num_cols * num_rows == 1:

			class MyC():
				def __init__(self, axis):
					self.axis = axis
				def ravel(self):
					return [self.axis]
			axeslist = MyC(axeslist)
		for ind,title in zip(range(len(names)), names):
			im = axeslist.ravel()[ind].imshow(logs_dict[title][1][-1,...], cmap='hot', vmin = _min, vmax = _max, interpolation='nearest', aspect='auto')
			axeslist.ravel()[ind].set_title(title)
			axeslist.ravel()[ind].set_axis_off()
		for last_idx in range(ind+1, num_cols*num_rows):
			axeslist.ravel()[last_idx].remove()
		plt.tight_layout() # optional

		fig.subplots_adjust(top=0.95,bottom=0.15, left=0.1, right=0.8)

		cbar_ax = fig.add_axes([0.85, 0.15, 0.05, 0.7])
		fig.colorbar(im, cax=cbar_ax)

		def update(val):
			for ind,title in zip(range(len(names)), names):
				if width_b.value_selected == 'original':
					im = axeslist.ravel()[ind].imshow(logs_dict[title][1][int((logs_dict[title][1].shape[0]-1)*samp.val),...], cmap=radio.value_selected, vmin = _min, vmax = _max)
				else:
					im = axeslist.ravel()[ind].imshow(logs_dict[title][1][int((logs_dict[title][1].shape[0]-1)*samp.val),...], cmap=radio.value_selected, vmin = _min, vmax = _max, interpolation='nearest', aspect='auto')

			fig.colorbar(im, cax=cbar_ax)

		axcolor = 'lightgoldenrodyellow'
		axsmooth = plt.axes([0.35, 0.025, 0.55, 0.03], facecolor=axcolor)
		samp = Slider(axsmooth, 'time', 0, 1, valinit=1)
		samp.on_changed(update)

		rax = plt.axes([0.025, 0.025, 0.1, 0.1], facecolor=axcolor)
		radio = RadioButtons(rax, ('hot', 'jet'), active=0)
		radio.on_clicked(update)

		rax2 = plt.axes([0.15, 0.025, 0.1, 0.1], facecolor=axcolor)
		width_b = RadioButtons(rax2, ('scaled', 'original'), active=0)
		width_b.on_clicked(update)

		plt.show()

def main(args=None):
	"""The main routine."""
	if args is None:
		args = sys.argv[1:]

	files_to_read = args

	for C in [Scalar, Histogram, Image]:
		c = C(files_to_read)


if __name__ == '__main__':
	main()



